
import xlrd
from contextlib import suppress


def convert(excel_file):
    for subclass in Converter.__subclasses__():
        converter = subclass(excel_file)

        return converter.convert()


class Converter:
    def __init__(self, excel_file):
        self._excel_file = excel_file
        self._file = xlrd.open_workbook(excel_file)

    def _get_columns(self, sheet_name, start=0):
        sheet = self._file.sheet_by_name(sheet_name)

        for colx in range(start, sheet.ncols):
            yield sheet.col_values(colx)

    def _get_capacities(self):
        raise NotImplementedError
    def _get_links(self):
        raise NotImplementedError
    def _get_network(self):
        raise NotImplementedError

    def convert(self):
        network_request = {
            'capacities': self._get_capacities(),
            'links': self._get_links(),
            'network': self._get_network()
        }

        return network_request


class NewFormatConverter(Converter):
    def _get_columns(self, sheet_name, start=0):
        return super()._get_columns(sheet_name, start=1)

    def _get_capacities(self):
        capacities = []
        for column in self._get_columns('Capacities'):
            (name, units, item_type, options, lower_bound, upper_bound) = column

            capacity = {
                'name': name,
                'units': units,
                'type': item_type
            }

            if lower_bound != '' or upper_bound != '':
                capacity['bounds'] = {}

                with suppress(ValueError):
                    capacity['bounds']['lower'] = int(lower_bound)
                with suppress(ValueError):
                    capacity['bounds']['upper'] = int(upper_bound)

            capacities.append(capacity)

        return capacities

    def _get_links(self):
        links = []
        for column in self._get_columns('Network links'):
            (link_id, start_id, end_id, link_type, length, capacity, voltage, electrical_resistance,
             electrical_reactance, total_thermal_loss, total_pressure_loss, operating_temperature) = column

            link ={
                'id': int(link_id),
                'start_id': int(start_id),
                'end_id': int(end_id),
                'type': link_type,
                # 'link_proportional_cost': link_proportional_cost, uncomment if want to set link cost in the excel file.
                # add 'link_cost' in the column names
                'length': length,
                'voltage': voltage,
                'resistance': electrical_resistance,
                'reactance': electrical_reactance,
                'total_thermal_loss': total_thermal_loss,
                'total_pressure_loss': total_pressure_loss,
                'operating_temp': operating_temperature,
            }

            try:
                capacity = float(capacity)
            except ValueError:
                capacity = str(capacity)

            link['capacity'] = capacity

            links.append(link)

        return links

    def _get_network(self):
        sheet = self._file.sheet_by_name('Network')

        return{
            'fixed_network_investment_cost': sheet.cell(0, 1).value,
            'link_proportional_cost': sheet.cell(1, 1).value, }








import openpyxl
from openpyxl import load_workbook
def main():

    book = openpyxl.load_workbook('Individual_Hubs/LI_Batresuls_11.xlsx')
    print(book.get_sheet_names())
    sheet = book.get_sheet_by_name('Other')

    # No of written Rows in sheet
    r = sheet.max_row
    # No of written Columns in sheet
    c = sheet.max_column
    # Reading each cell in excel
    for i in range(1, r+1):
        for j in range(1, c+1):
            print(sheet.cell(row=i, column=j).value)








if __name__ == '__main__':
    main()


import collections
from collections import OrderedDict
from typing import Union
import yaml
import pandas as pd
from openpyxl import load_workbook
from energy_hub.ehub_model import CONFIG_PATH

EXCEL_ENGINE = 'openpyxl'
EXCEL_FIRST_COLUMN_WIDTH = 25


""" Loads a yaml file """
with open(CONFIG_PATH, "r") as file_descriptor:
    data = yaml.load(file_descriptor)


def to_dataframes(frames: dict) -> OrderedDict:
    """
    Convert the values of a dictionary into dataframes if they can be converted.

    Args:
        frames: The dictionary to be converted

    Returns:
        An ordered dictionary with dataframes as values (if they can be
        converted).
    """
    for name, value in frames.items():
        if isinstance(value, dict):
            value = to_dataframe(name, value)

        frames[name] = value

    return OrderedDict(sorted(frames.items()))


def to_dataframe(name: str, value: dict) -> pd.DataFrame:
    """
    Convert a dictionary into a dataframe.

    Args:
        name: The name for the dataframe
        value: The dictionary to be converted

    Returns:
        A Pandas DataFrame.
    """
    value = sort_dict(value)
    value = pd.DataFrame.from_dict(value, orient='index')

    # To remove confusion on what the column '0' means
    if list(value.columns) == [0]:
        value.columns = [name]

    # Make wide matrices fit on the screen
    num_rows, num_columns = value.shape
    if num_columns > num_rows:
        value = value.T

    return value


def _create_sheet(excel_writer: pd.ExcelWriter, sheet_name: str):
    """
    Create an Excel sheet with the given name.

    Args:
        excel_writer: The ExcelWriter object
        sheet_name: The name of the new sheet

    Returns:
        The worksheet for the new sheet
    """
    # Create an empty dataframe just so that we can create a sheet
    pd.DataFrame().to_excel(excel_writer, sheet_name)

    return excel_writer.sheets[sheet_name]


def output_excel(solution: dict, file_name: str, time_steps: int = None, sheets:list = None):
    """ 
    Output the solution dictionary to an Excel file.

    It outputs the time series data in their own sheet with the rest being put
    into another sheet.

    Args:
        solution: A dictionary of the solution part of the response format.
            This contains the variables and parameters of the solved model.
        file_name: The name of the Excel file to write to
        time_steps: Optional. The number of time steps to classify a dataframe
            as holding time series data.
        sheets: A list of all the sheets to be contained in the excel file.
    """
    excel_writer = pd.ExcelWriter(file_name, engine=EXCEL_ENGINE)
    attributes = to_dataframes(solution)

    def _has_time_series_data(value):
        if isinstance(value, pd.DataFrame):
            num_rows, _ = value.shape

            return time_steps and num_rows >= time_steps
        return False

    time_series = {name: value for name, value in attributes.items()
                   if _has_time_series_data(value)}
    non_time_series = {name: value for name, value in attributes.items()
                       if name not in time_series}

    for name, value in time_series.items():
        value.to_excel(excel_writer, f'{name}')

    #: Set first_sheet to "Other" if sheets is not passed, otherwise set it to
    #: the first item in sheets.
    first_sheet = 'Other' if sheets is None else sheets[0]

    #: If sheets is passed, then pass all the sheets except the first as additional sheets.
    if sheets is not None:
        _dict_to_excel_sheet(excel_writer, first_sheet, non_time_series, padding=1, additional_sheets= sheets[1:])
    else:
        #: Otherwise dont sent any value for additional sheets, as none have been provided.
        _dict_to_excel_sheet(excel_writer, first_sheet, non_time_series, padding=1)

    excel_writer.save()


def _dict_to_excel_sheet(excel_writer, sheet_name, non_time_series, padding=0, additional_sheets: list=None):
    """ 
    Output a dictionary to an excel sheet.

    Args:
        excel_writer: The ExcelWriter object
        sheet_name: The name of the sheet to output to
        non_time_series: The dictionary of data
        padding: The padding between key, value pairs
        additional_sheets: a list containing names of other sheets
                           to potentially be added to the excel document.
    """
    hold_row = 1
    row = 1
    hold_sheet_name = sheet_name
    sheet = _create_sheet(excel_writer, sheet_name)
    hold_sheet = sheet

    for name, value in non_time_series.items():
        if additional_sheets is not None and name in additional_sheets:
            hold_row = row
            row = 1
            sheet = _create_sheet(excel_writer, name)
            sheet_name = name
        if isinstance(value, pd.DataFrame):
            num_rows, num_cols = value.shape

            # Compensate for empty Excel row of dataframes. No idea why. Gee thanks
            # Also handle case for when printing first value in sheet.
            startrow = max(row - 1, 1)
            value.to_excel(excel_writer, sheet_name, startrow=startrow)

            if num_cols > 1:
                # Output variable name if not in column header
                sheet.cell(row=row, column=1).value = name

            row += num_rows + 1  # Add row for column headers
        else:
            sheet.cell(row=row, column=1).value = name

            if not isinstance(value, collections.Iterable):
                # Simplify writing column values
                value = [value]

            for col, x in enumerate(value, start=2):
                sheet.cell(row=row, column=col).value = str(x)

            row += 1

        if sheet_name != hold_sheet_name:
            sheet_name = hold_sheet_name
            row = hold_row
            sheet = hold_sheet

        row += padding

    sheet.column_dimensions['A'].width = EXCEL_FIRST_COLUMN_WIDTH


def pretty_print(results: dict) -> None:
    """Print the results in a prettier format.

    Args:
        results: The results dictionary to print
    """
    version = results['version']
    solver = results['solver']

    print("Version: {}".format(version))

    print("Solver")
    print(f"\ttermination condition: {solver['termination_condition']}")
    if solver['termination_condition']!='Infeasible':
        print(f"\ttime: {solver['time']}")

        print("Solution")
        print_section('Stuff', results['solution'])


def print_section(section_name: str, solution_section: dict) -> None:
    """
    Print all the attributes with a heading.

    Args:
        section_name: The heading
        solution_section: The dictionary with all the attributes
    """
    half_heading = '=' * 10
    print(f"\n{half_heading} {section_name} {half_heading}")

    attributes = to_dataframes(solution_section)
    for name, value in attributes.items():
        with pd.option_context('display.max_rows', None, 'display.max_columns', 3):
            print(f"\n{name}: \n{value}")


def sort_dict(mapping: Union[dict, OrderedDict]) -> OrderedDict:
    """
    Sorts a dictionary and all its sub-dicionaries as well.

    Examples:
        >>> sort_dict({1: 'a', 3: 'c', 2: 'b'})
        OrderedDict([(1, 'a'), (2, 'b'), (3, 'c')])
        >>> sort_dict({1: {3: 'c', 2: 'b', 1: 'a'}})
        OrderedDict([(1, OrderedDict([(1, 'a'), (2, 'b'), (3, 'c')]))])

    Args:
        mapping: The dictionary to be sorted

    Returns:
        The sorted dictionary as an OrderedDict
    """
    if not isinstance(mapping, dict):
        return mapping

    for key, value in mapping.items():
        mapping[key] = sort_dict(value)

    return OrderedDict(sorted(mapping.items()))


def print_capacities(results):
    """
    Prints the capacities of each tech and storage at the end

    :param results: the solved model

    """
    solution = results['solution']
    capacity_tech = solution['capacity_tech']
    capacity_storage = solution['capacity_storage']

    half_heading = '=' * 10
    print(f"\n{half_heading} {'Capacities'} {half_heading}")
    print(half_heading * 4)
    print(capacity_storage)
    print(half_heading * 4)
    print(capacity_tech)


def print_warning(results):
    """
    Prints an error if the model burns energy, i.e there is energy from storage and energy to storage
    at same time step.

    :param results: the solved model

    """
    solution = results['solution']
    storages = solution['storages']
    energy_from_storage = solution['energy_from_storage']
    energy_to_storage = solution['energy_to_storage']
    time = len(solution['time'])

    half_heading = '=' * 4

    for storage in storages:
        for t in range(time):
            energy_from = energy_from_storage[storage][t]
            energy_to = energy_to_storage[storage][t]

            if (energy_to != 0) and (energy_from != 0):
                print(f"{half_heading} {'Warning:'} {half_heading} \n{ 'Burning energy for storage: '} {storage} "
                      f"{'; For time step: '} {t} ")


def stream_info(results, output_file):
    """
    New output format with the information separated in different sheets for different streams.

    :param results: the solved model
    :param output_file: the output excel file
    """

    solution = results['solution']
    streams = solution['streams']
    load = solution['LOADS']
    energy_from_storage = solution['energy_from_storage']
    energy_to_storage = solution['energy_to_storage']
    energy_exported = solution['energy_exported']
    energy_imported = solution['energy_imported']
    demands = solution['demands']
    time = len(solution['time'])
    export_streams = solution['export_streams']
    import_streams = solution['import_streams']
    storages = solution['storages']
    storage_level = solution['storage_level']

    n = len(storages)


    with pd.ExcelWriter(output_file, engine=EXCEL_ENGINE) as writer:
        writer.book = load_workbook(output_file)

    keep = ['other', 'TIME_SERIES', 'energy_input', 'capacity_tech', 'capacity_storage']
    for sheet_name in writer.book.sheetnames:
        if sheet_name not in keep:
            del writer.book[sheet_name]

    for stream in streams:

        if stream in demands:
            df = pd.DataFrame({
                stream + " Load": [load[stream][t] for t in range(time)]})
            df.to_excel(writer, sheet_name=stream, startrow=time + 2)

        df1 = pd.DataFrame({
            "energy_from " + storage: [energy_from_storage[storage][t] for t in range(time)] for storage in storages})

        df2 = pd.DataFrame({
            "energy_to " + storage:  [energy_to_storage[storage][t] for t in range(time)] for storage in storages})

        df3 = pd.DataFrame({
            "level " + storage: [storage_level[storage][t] for t in range(time)] for storage in storages})

        df1.to_excel(writer, sheet_name=stream, )
        df2.to_excel(writer, sheet_name=stream, startcol=n + 2)
        df3.to_excel(writer, sheet_name=stream, startcol=2*(n + 2))

        if stream in import_streams:
            df4 = pd.DataFrame({
                stream + " imported": [energy_imported[stream][t] for t in range(time)]})
            df4.to_excel(writer, sheet_name=stream, startrow=time + 2, startcol=4)

        if stream in export_streams:
            df5 = pd.DataFrame({
                stream + " exported": [energy_exported[stream][t] for t in range(time)]})
            df5.to_excel(writer, sheet_name=stream, startrow=time + 2, startcol=7)

        writer.save()



import argparse
import os
import re
import shutil
import openpyxl
from openpyxl import Workbook
from openpyxl.styles import Alignment
from docs.tutorials.multiple_hubs import multiple_hubs

""" File naming convention:

In order for an excel file to be read as input correctly, it must
follow the follwing naming convention:

[any valid file name]_[id to be assosiated with it].xlsx

note,
	carbon/cost is there to allow you to specify whether a test is trying to minimize
	either cost or carbon.

"""

def get_output_fields(filenames: list, directory: str = None):
	""" get_output_fields(list, str)

		Gets the required outputs from the excel files passed in the
		list of filenames.

		Args:
			filenames(list): A list of the excel files to find outputs for.

			directory(str): The optional dicretory that these filenames are located at.
					for use only if the filenames do not include their directory.

		Returns:
			A list with every output requested from every file (duplicates are deleted.)
	"""

	#: Create the list to hold the values in.
	master_field_list = []

	for filename in filenames:

		#: Open the excel file for this loop.
		if directory is not None:
			book = openpyxl.load_workbook(directory + filename)
		else:
			book = openpyxl.load_workbook(filename)

		#: Open the capacities sheet which is set to have all of
		#: the required outputs in it.
		sheet = book["Capacities"]

		#: Loop through the header (skipping not output fields)\
		#: and get all of the required outputs. These exist from
		#: (row 1, col 2) to (row 1, final col)
		fields = sheet.iter_rows(min_row=1, max_row=1, min_col=2)

		#: Convert the generator to a list, and get the only row.
		#: then loop through each cell in the row, and append the
		#: value in that cell to the master list.
		for cell in list(fields)[0]:
			master_field_list.append(cell.value)

	#: Append total_cost and total_cost to the output list so that these
	#: values are automatically included in the output.
	master_field_list.append("total_cost")
	master_field_list.append("total_carbon")
	#: Convert the list to a set to remove all duplicates,
	#: then convert the set back to a list, and return.
	return_list = []
	for item in master_field_list:
		if item is not None:
			return_list.append(item)

	return sorted(list(set(return_list)))


def output_to_excel(outputs: dict, sheet, output_dict: dict):
	""" output_to_excel(dict, openpyxl.sheet, dict)

		Takes the output dictionary produced by multiple hubs, and
		writes specific fields to the passed excel sheet based on
		the output dictionary passed.

		Args:
			outputs(dict): The dictionary produced by multiple_hubs.

			sheet(openpyxl.sheet): The excel sheet to be written to.

			output_dict(dict): A dictionary that links keys found in the
					   outputs dict, and the column that they should
					   appear in, within the sheet.

	"""

	#: Get all of the columns that have content in them, and use
	#: that to figure out where the next blank column is.
	cols = list(sheet.iter_cols(min_col=1, max_col=1))[0]
	first_empty_row = len(cols) + 1


	#: Merge the cells that have large words togther. These are the
	#: first 4 cells appearing in any output (Filename, funcion,
	#: fjob, and hub).
	#: File name header cell
	sheet.merge_cells("A{}:B{}".format(first_empty_row, first_empty_row))
	#: Function header cell
	sheet.merge_cells("C{}:D{}".format(first_empty_row, first_empty_row))
	#: fjob header cell
	sheet.merge_cells("E{}:F{}".format(first_empty_row, first_empty_row))
	#: Hub header cell
	sheet.merge_cells("G{}:H{}".format(first_empty_row, first_empty_row))

	#: for each key and value pair in the formatting dictionary passed.
	for key, value in output_dict.items():

		#: If outputs[key] exists (in other words if the current key from output_dict,
		#: represents an output that is in this set of outputs, then set it to the corrisponding
		#: column based off the value of this item.
		try:
			sheet['{}{}'.format(value, first_empty_row)] = outputs[key]

		#: Otherwise move on to the next loop, as this output was not outputted by this file.
		except:
			continue

def main(batch_run_location: str="excel_files/batch_input/", output_directory: str="batchrun_outputs/", epsilon_n: int = 0):

	#: If argparse got no -d switch
	if output_directory is None:

		#:Set output directory to its default setting.
		output_directory = "batchrun_outputs/"

	#: Otherwise, if the user forgot to / or \ terminate their directory, then
	#: do it for them.
	elif output_directory[-1] != "/" and output_directory[-1] != "\\":
		output_directory = output_directory + "/"

	#: If argparse got no -o switch
	if batch_run_location is None:

		#: Set batch_run_location to its default setting.
		batch_run_location = "excel_files/batch_input/"

	#: Otherwise if the user forgot to / or \ terminate their directory, then
	#: do it for them.
	elif batch_run_location[-1] != "/" and batch_run_location[-1] != "\\":
		batch_run_location = batch_run_location + "/"

	#: If argparse got no -n switch
	if epsilon_n is None:

		#: Set epsilon_n to iss default setting.
		epsilon_n = 0

	#: If the output directory is not currently a directory, make it.
	if not os.path.exists(output_directory):
		os.makedirs(output_directory)

	#: Get a list of all files in the run location that end in .xlsx (excel's file
	#: extension)
	excel_files = [file for file in os.listdir(batch_run_location) if file[-5:] == ".xlsx" and file[0] != '$']

	file_data = [] #: An intermediate container to hold some of the information
		       #: That will be gathered and held on the files.

	#: Get the id of every file found in the run location,
	#: then store a tuple with (filename, fileid) into file_data.
	for file in excel_files:

		#: Initial position while looking for the _.
		_position = 0

		#: For each character in the file name (backwards),
		#: if the character is an underscore then we know that
		#: this is the character before the id starts.
		for character in file[::-1]:

			#: Check if the character is an _, if so then
			#: we're done looking and should break from the
			#: inner loop.
			if character == "_":
				break

			#: Otherwise subtract 1 from position.
			_position -= 1

		#: Append a tuple defined by (filename, fileid).
		#: file[_position:-5] gets the id because _position
		#: is the location of the LAST _ in the filename (ie, the
		#: one that is required to preceed the id) and there are five
		#: characters at the end of each file ".xlsx". Thus the range
		#: between _position, and 5 is the id (no matter how many numbers.)
		file_data.append((file, file[_position:-5]))

	#: Sort file data based off of the file ids so that
	#: all pairs of files are beside eachother in the list.
	file_data = sorted(file_data, key=lambda x: x[1])

	file_sets = [] #: The final container to hold tuples of pairs of filenames.

	#: Append a tuple defined by (filename1, filename1's pair) to file_sets.
	#: for each pair in file_data.
	for index, file in enumerate(file_data[::2]):
		file_sets.append((file, file_data[(index * 2) + 1]))

	#: Get all the required outputs from the found excel files.
	all_fields = get_output_fields(excel_files, directory=batch_run_location)

	#: The first column to put data in is alwasy I
	first_letter = 'I'

	#: The number of fields automatically appeneded after the capacities.
	#: These fields are:
	#:	total_carbon
	#:	total_cost
	extra_fields = 2

	#: Calculate the letter(s) of the final column to put data in. - 1 is to
	#: account for an off by one error. The complexity in these functions is
	#: to account for the ability of the sheet to be over 26 columns long. In
	#: that case, excel starts using AA, then AB, etc so this must be accounted for.
	final_letter = str(first_letter)
	for _ in range(len(all_fields) - 1 - extra_fields):
		#: If the final letter is not a Z, then just increase it by a letter.
		if final_letter[-1] != 'Z':
			final_letter = final_letter[:-1] + chr(ord(final_letter[-1]) + 1)
		#: Otherwise remove the Z and replace it with AA to account for
		#: excel's column naming.
		else:
			final_letter = final_letter[:-1] + "AA"

	#: Create a list of the extended row's letter(s) representing their columns.
	extended = [final_letter]
	for _ in range(extra_fields):
		if extended[-1][-1] != 'Z':
			extended.append(extended[-1][:-1] + chr(ord(extended[-1][-1]) + 1))
		else:
			extended.append(extended[-1][:-1] + "AA")

	#: Create an excel workbook, and open a sheet.
	output_book = Workbook()
	sheet = output_book.active

	#: Merge cells for proper formatting of headers.
	#: Filename header
	sheet.merge_cells("A1:B2")
	#: Function header
	sheet.merge_cells("C1:D2")
	#: fjob header
	sheet.merge_cells("E1:F2")
	#: Hub header
	sheet.merge_cells("G1:H2")
	#: Capacity header: must span from first output to last output.
	sheet.merge_cells("{}1:{}1".format(first_letter, final_letter))
	#: Totals header must span from final_letter + 1 to final_letter + 2
	sheet.merge_cells("{}1:{}1".format(extended[1], extended[2]))

	#: Assign header values to these cells.
	#: Filename header
	sheet['A1'] = "File Name"
	#: Function header
	sheet['C1'] = "Function"
	#: fjob header
	sheet['E1'] = "fjob"
	#: Hub header
	sheet['G1'] = "Hub"
	#: Use first_letter here instead of 'I' to increase readability.
	sheet['{}1'.format(first_letter)] = "Capacity"
	#: Totals header
	sheet['{}1'.format(extended[1])] = "Totals"

	#: A dictionary describing which column data for each header should
	#: go in. Used as the output_dict parameter for output_to_excel().
	pos_dict = {
		     #: header: Data column,
			"Filename": "A",
			"Function": "C",
			"fjob": "E",
			"Hub": "G"
		}
	#: Create a string to parse through the data columnsm
	#: and to set values with.
	fletter = str(first_letter)

	#: For each field, append that field's name as a subheader under
	#: capacity.
	for field in all_fields:

		#: Append the current field name to the column
		#: stored currently in fletter.
		sheet['{}2'.format(fletter)] = field

		#: Build the dictionary used in output_to_excel().
		#: set the key value of field to the current letter
		#: stored by fletter.
		pos_dict[field] = fletter

		#: Increment fletter by one. If fletter's last (only) character
		#: is Z, then change it to AA
		if fletter[-1] != 'Z':
			fletter = fletter[:-1] + chr(ord(fletter[-1]) + 1)
		else:
			fletter = fletter[:-1] + "AA"

	#: Centre the cells all of the cells that we have used as headers.
	#: The Filename header
	sheet['A1'].alignment = Alignment(horizontal='center', vertical='center')
	#: The function header
	sheet['C1'].alignment = Alignment(horizontal='center', vertical='center')
	#: The fjob header
	sheet['E1'].alignment = Alignment(horizontal='center', vertical='center')
	#: The Hub header.
	sheet['G1'].alignment = Alignment(horizontal='center', vertical='center')
	#: Center the merged capacity cell.
	sheet['{}1'.format(first_letter)].alignment = Alignment(horizontal='center', vertical='center')
	#: Center the merged Totals cell.
	sheet['{}1'.format(extended[1])].alignment = Alignment(horizontal='center', vertical='center')

	#: Main execution loop. This handles the execution of each pair of files that was found.
	#: it also handles all of the outputting of the data found by the executions. Finally, it
	#: also handles looping over different carbon levels if epsilon_n is not 0.
	for fjob, file_set in enumerate(file_sets):

		#: Load the data from the first two excel files into the hub1 and hub2 files as required
		#: multiple_hubs.py
		shutil.copyfile(batch_run_location + file_set[0][0], 'docs/tutorials/network/hub1.xlsx')
		shutil.copyfile(batch_run_location + file_set[1][0], 'docs/tutorials/network/hub2.xlsx')

		#: Output current action as some actions may take a while and we want the user to know
		#: that we're not just frozen.
		print("Finding minimum cost for {} and {}".format(file_set[0][0][:-5], file_set[1][0][:-5]))

		#: Attempt to minimize the cost without worrying about the carbon output. Store the
		#: optimized data into cost_min.
		cost_min = multiple_hubs(False, output_directory + file_set[0][0][:_position - 1])

		#: Get which outputs are given by hub1 and hub2. NOTE these aren't the outputs,
		#: they are just the fields of important outputs to be logged in final_results.xlsx
		hub1_output_fields = get_output_fields([file_set[0][0]], directory=batch_run_location)
		hub2_output_fields = get_output_fields([file_set[1][0]], directory=batch_run_location)

		#: Create the base dictionary to output data with. This dictionary will
		#: be used as the outputs variable for output_to_excel().
		#: This base dictionary creates the information that isnt returned by
		#: multiple_hubs; this is mostly just readability data (ie used to distinguish
		#: between different runs.)
		hub1_outputs = {
				"Filename": file_set[0][0][:-5],
				"Function": "Minimize Cost",
				"fjob": (2 * fjob) + 1,
				"Hub": 1
			}

		#: Create the base dictionary to output data with. This dictionary will
		#: be used as the outputs variable for output_to_excel().
		#: This base dictionary creates the information that isnt returned by
		#: multiple_hubs; this is mostly just readability data (ie used to distinguish
		#: between different runs.)
		hub2_outputs = {
				"Filename": file_set[1][0][:-5],
				"Function": "Minimize Cost",
				"fjob": (2 * fjob) + 1,
				"Hub": 2
			}

		#: For each field that we need to store for hub 1,
		#: add an entry to hub1_outputs where the key is
		#: field, and the value is the value for that field
		#: found by running multiple_hubs()
		for field in hub1_output_fields:
			hub1_outputs[field] = cost_min[0][field]

		#: For each field that we need to store for hub 2,
		#: add an entry to hub1_outputs where the key is
		#: field, and the value is the value for that field
		#: found by running multiple_hubs()
		for field in hub2_output_fields:
			hub2_outputs[field] = cost_min[1][field]

		#: Output all of the data found to the working excel file.
		output_to_excel(outputs=hub1_outputs, sheet=sheet, output_dict=pos_dict)
		output_to_excel(outputs=hub2_outputs, sheet=sheet, output_dict=pos_dict)

		#: Find the average carbon used on a run that minimizes cost (and theoretically
		#: maximizes carbon.)
		avg_carbon = (cost_min[0]["total_carbon"] + cost_min[1]["total_carbon"]) / 2

		#: Find the value of the carbon that should be used as a maximum when
		#: multiplied by the current step.
		carbon_per_step = 0 if epsilon_n == 0 else avg_carbon / epsilon_n

		#: Handles the epsilon function.
		#: If epsilon_n is 0, then this will only run once in an attempt to
		#: minimize cost at certain (n * carbon_per_step) carbon maximums.
		#Otherwise, we will run espilon_n times + 1 for the minimize carbon run.
		for n in range(epsilon_n + 1):

			#: Print which files and what n value we're currently at (also print
			#: the maximum carbon for this run.)
			print("Running {} and {}  at n = {}, max carbon = {}"\
				.format(file_set[0][0][:-5], file_set[1][0][:-5], n, carbon_per_step * n))

			#: Attempt to minimize the cost, with a maximum carbon value of carbon_per_step * n.
			#: Note,
			#:	If this is the first/only loop then carbon will be maxed at zero, while attempting
			#:	to optimize for cost. This will simulate a no-carbon building, and will theoretically
			#:	be the maximum cost. On all other runs, the carbon will be maxed at carbon_per_step * n.
			carbon_min = multiple_hubs(True, output_directory + "n-{}".format(n) + file_set[0][0][:_position - 1],
						   carbon_per_step * n)

			#: Create the base dictionary to output data with. This dictionary will
			#: be used as the outputs variable for output_to_excel().
			#: This base dictionary creates the information that isnt returned by
			#: multiple_hubs; this is mostly just readability data (ie used to distinguish
			#: between different runs.)
			hub1_outputs = {
					"Filename": file_set[0][0][:-5],
					"Function": "Minimize Carbon; n = {}".format(n),
					"fjob": (fjob + 1) * 2,
					"Hub": 1
				}

			#: Create the base dictionary to output data with. This dictionary will
			#: be used as the outputs variable for output_to_excel().
			#: This base dictionary creates the information that isnt returned by
			#: multiple_hubs; this is mostly just readability data (ie used to distinguish
			#: between different runs.)
			hub2_outputs = {
					"Filename": file_set[1][0][:-5],
					"Function": "Minimize Carbon; n = {}".format(n),
					"fjob": (fjob + 1) * 2,
					"Hub": 2
				}

			#: For each field that we need to store for hub 1,
			#: add an entry to hub1_outputs where the key is
			#: field, and the value is the value for that field
			#: found by running multiple_hubs()
			for field in hub1_output_fields:
				hub1_outputs[field] = carbon_min[0][field]

			#: For each field that we need to store for hub 2,
			#: add an entry to hub1_outputs where the key is
			#: field, and the value is the value for that field
			#: found by running multiple_hubs()
			for field in hub2_output_fields:
				hub2_outputs[field] = carbon_min[1][field]

			#: Output all of the data found to the working excel file.
			output_to_excel(outputs=hub1_outputs, sheet=sheet, output_dict=pos_dict)
			output_to_excel(outputs=hub2_outputs, sheet=sheet, output_dict=pos_dict)

	#: Save the excel spread sheet that was used to store all of the runs's important outputs.
	output_book.save(output_directory + "final_results.xlsx")

if __name__ == "__main__":

	#: Create the argument parser to parse the passed arguments.
	parser = argparse.ArgumentParser(description="A commandline utility to batch run tests.")

	#: Add the -d argument.
	parser.add_argument('-d', '--directory_name', type=str, help='The directory containing the excel file to be run.\n'\
									'Default is excel_files/batch_input/')
	#: Add the -o argument.
	parser.add_argument('-o', '--output_name', type=str, help='The directory to store all outputted files in..\n'\
									'Default is batchrun_outputs/')
	#: Add the -n argument.
	parser.add_argument('-n', '--epsilon_n', type=int, help='The number of steps to break into for the epsilon constraint.')

	#: Parse the passed arguments.
	args = parser.parse_args()

	#: Pass the arguments that were passed.
	main(args.directory_name, args.output_name, args.epsilon_n)


"""
Solving energy hub model for n number of hubs with any network of connections wanted between the hubs.

To run for multiple hubs:
    $ python multiple_hubs.py -n NUMBER OF HUBS

If n (NUMBER OF HUBS) is not inputted the code will not run.

Naming input excel files:
    file names should start with "hub" followed by increasing numbers starting from 0.
    The files should be in the 'network' folder

To set the connections between the hubs ( the links):
   All the links between the hubs should be set in a separate excel file in the 'network' folder.
   It should be called "network.xlsx"

   The default set up is one directional connections.
   To set bidirectional connections, the connections should be defined per each direction, i.e:

        one direction:  link 0:  start_id: 0   end_id: 1


        bidirectional:  link 0:  start_id: 0   end_id: 1,
                        link 1:  start_id: 1   end_id: 0

    The link ids should start from 0 and increase by 1.
    The node ids (start_id and end_id) correspond to hubs numbering in the names of the hubs excel files.

Note:
Do not name constraints specific names in the EHubModel class -> will not be able to construct
                                                                 constraints in multiple_hubs
"""
import os
import pylp
import argparse

from pylp import RealVariable, BinaryVariable, IntegerVariable
from energy_hub.param_var import ConstantOrVar
from energy_hub import EHubModel
from energy_hub.utils import constraint
from outputter import print_section, output_excel
import network_to_request_format
from energy_hub.input_data import InputData
import numpy as np


DOMAIN_TO_VARIABLE = {
    'Continuous': RealVariable,
    'Integer': IntegerVariable,
    'Binary': BinaryVariable,
}


class NetworkModelWithTotalCarbon(EHubModel):
    """
    A subclass that allows connections between hubs.
    """

    MAX_CARBON = 0

    def __init__(self, *, excel=None, request=None, name=None, n=0, network=None, hub_id=None):
        super().__init__(excel=excel, request=request)
        if n is not None:
            self.MAX_CARBON = n
        self.name = name
        self.hub_id = hub_id
        network_request = network_to_request_format.convert(network)
        self._net_data = InputData(network_request)

        self.links = self._net_data.links_ids
        self.link_capacities = ConstantOrVar(self.links, model=self, values=self._net_data.link_capacity)
        self.network_cost = RealVariable()
        self.is_link_installed = {link: BinaryVariable()
                                  for link in self.links}

        self.LINK_LENGTH = self._net_data.link_length
        self.FIXED_NETWORK_INVESTMENT_COST = self._net_data.fixed_network_investment_cost
        self.LINK_PROPORTIONAL_COST = self._net_data.link_proportional_cost

    def _add_variables(self):
        super()._add_variables()
        self._add_link_capacity_variables()
        for i in self._net_data.capacities:
            self._data.capacities.append(i)

    @constraint('links')
    def link_is_installed(self, link):
        """
        Set binary to 1 if capacity of link is > 0.
        Args:
            link: A link
        """
        capacity = self.link_capacities[link]
        rhs = self.BIG_M * self.is_link_installed[link]
        return capacity <= rhs

    @constraint('links')
    def link_is_installed_2(self, link):
        """
        Set binary to 1 if capacity of link is > 0.
        Args:
            link: A link
        """
        installed = self.is_link_installed[link]
        lhs = self.BIG_M * self.link_capacities[link]
        return lhs >= installed

    # Calculating investment cost for the links the hub has.
    # Cost split between 2 hubs that are connected with the same link
    @constraint()
    def calc_network_investment_cost(self):
        cost = 0
        for i, link in enumerate(self.links):
            if (self._net_data.link_start[i] == self.hub_id) or (self._net_data.link_end[i] == self.hub_id):
                cost += (self.FIXED_NETWORK_INVESTMENT_COST * self.LINK_LENGTH[link] * self.is_link_installed[link]
                         + self.link_capacities[link] * self.LINK_LENGTH[link] * self.LINK_PROPORTIONAL_COST)

        cost = cost / 2
        return self.network_cost == cost

    # Modifying total cost constraint. Adding network cost to the total cost.
    @constraint()
    def calc_total_cost(self):
        parent_constraint = super().calc_total_cost()
        old_cost = parent_constraint.rhs

        return self.total_cost == old_cost + self.network_cost

    def _add_link_capacity_variables(self):
        for capacity in self._net_data.capacities:
            domain = capacity.domain
            name = capacity.name

            try:
                variable = DOMAIN_TO_VARIABLE[domain]()
            except KeyError:
                raise ValueError(
                    f'Cannot create variable of type: {domain}. Can only be: '
                    f'{list(DOMAIN_TO_VARIABLE.keys())}'
                )

            setattr(self, name, variable)

    # A constraint for epsilon constraint method
    @constraint()
    def max_carbon_level(self):
        return self.total_carbon <= float(self.MAX_CARBON)


class NetworkModelWithoutTotalCarbon(EHubModel):
    """
    A subclass that allows connections between hubs.
    """

    def __init__(self, *, excel=None, request=None, name=None, network=None, hub_id=None):
        super().__init__(excel=excel, request=request)
        self.name = name
        self.hub_id = hub_id
        network_request = network_to_request_format.convert(network)
        self._net_data = InputData(network_request)

        self.links = self._net_data.links_ids
        self.link_capacities = ConstantOrVar(self.links, model=self, values=self._net_data.link_capacity)
        self.network_cost = RealVariable()
        self.is_link_installed = {link: BinaryVariable()
                                  for link in self.links}

        self.LINK_LENGTH = self._net_data.link_length
        self.FIXED_NETWORK_INVESTMENT_COST = self._net_data.fixed_network_investment_cost
        self.LINK_PROPORTIONAL_COST = self._net_data.link_proportional_cost

    def _add_variables(self):
        super()._add_variables()
        self._add_link_capacity_variables()
        for i in self._net_data.capacities:
            self._data.capacities.append(i)

    @constraint('links')
    def link_is_installed(self, link):
        """
        Set binary to 1 if capacity of link is > 0.
        Args:
            link: A link
        """
        capacity = self.link_capacities[link]
        rhs = self.BIG_M * self.is_link_installed[link]
        return capacity <= rhs

    @constraint('links')
    def link_is_installed_2(self, link):
        """
        Set binary to 1 if capacity of link is > 0.
        Args:
            link: A link
        """
        installed = self.is_link_installed[link]
        lhs = self.BIG_M * self.link_capacities[link]
        return lhs >= installed

    # Calculating investment cost for the links the hub has.
    # Cost split between 2 hubs that are connected with the same link
    @constraint()
    def calc_network_investment_cost(self):
        cost = 0
        for i, link in enumerate(self.links):
            if (self._net_data.link_start[i] == self.hub_id) or (self._net_data.link_end[i] == self.hub_id):
                cost += (self.FIXED_NETWORK_INVESTMENT_COST * self.LINK_LENGTH[link] * self.is_link_installed[link]
                         + self.link_capacities[link] * self.LINK_LENGTH[link] * self.LINK_PROPORTIONAL_COST)

        cost = cost/2
        return self.network_cost == cost

    # Modifying total cost constraint. Adding network cost to the total cost.
    @constraint()
    def calc_total_cost(self):
        parent_constraint = super().calc_total_cost()
        old_cost = parent_constraint.rhs

        return self.total_cost == old_cost + self.network_cost

    def _add_link_capacity_variables(self):
        for capacity in self._net_data.capacities:
            domain = capacity.domain
            name = capacity.name

            try:
                variable = DOMAIN_TO_VARIABLE[domain]()
            except KeyError:
                raise ValueError(
                    f'Cannot create variable of type: {domain}. Can only be: '
                    f'{list(DOMAIN_TO_VARIABLE.keys())}'
                )

            setattr(self, name, variable)


@constraint()
def network_constraint(hub, link_end, link_start):
    """
    Yields the constraints that allow a network connection between two hubs.

    Args:
        hub: The hub
        link_end: all the links that the hub ends at
        link_start: all the links that the hub start at
    Yields:
       A network energy balanced constraints for each hub
    """

    for t in hub.time:
        link_starting = []
        link_ending = []

        for i in range(len(link_start)):
            link_starting.append(link_start[i][t])

        for i in range(len(link_end)):
            link_ending.append(link_end[i][t])

        yield ((hub.energy_exported[t]['Net_export'] - hub.energy_imported[t]['Net_import'] + sum(link_ending)
                - sum(link_starting)) == 0)


# constraint for the flow in the links
@constraint()
def link_capacity_constraint(link, hub, i):
    for flow in link:
        yield flow >= 0
        yield flow <= hub.link_capacities[i]


def multiple_hubs(minimize_carbon=False, output_filename=None, input_files=None, network_excel=None, carbon_value=0, n=0):
    if carbon_value is None:
        carbon_value = 0

    # The default excel file for network, in the network directory
    if network_excel is None:
        network_excel_file = 'Input_files/network_input/network.xlsx'
    # Additional option for tests for multiple_hubs. To input a different network excel files.
    else:
        network_excel_file = network_excel
    network = network_to_request_format.convert(network_excel_file)
    _net_data = InputData(network)

    # Create all the hubs
    hubs = []
    for i in range(0, n):
        # Default files in the network directory
        if input_files is None:
            file_name = f'Input_files/network_input/hub{i + 1}.xlsx'
        # Additional options for test for multiple hubs. To input a different directory and excel files
        else:
            file_name = f'{input_files}{i+1}.xlsx'
        # This is a cross-platform way of getting the path to the Excel file
        current_directory = os.path.dirname(os.path.realpath(__file__))
        excel_file = os.path.join(current_directory, file_name)

        if minimize_carbon:
            hub = NetworkModelWithTotalCarbon(excel=excel_file, name=f'hub{i+1}', n=carbon_value,
                                              network=network_excel_file, hub_id=i)
        else:
            hub = NetworkModelWithoutTotalCarbon(excel=excel_file, name=f'hub{i+1}',
                                                 network=network_excel_file, hub_id=i)
        hubs.append(hub)

    # Creating a list of all the constraints from all the hubs
    constraints = []
    for hub in hubs:
        hub.recompile()
        for constr in hub.constraints:
            constraints.append(constr)

    # The connections between the hubs, from the excel network input file.
    connections = []
    for i in _net_data.links_ids:
        start = _net_data.link_start[i]
        end = _net_data.link_end[i]

        connections.insert(i, (start, end))

    print(connections)
    links = []

    # The energy flow in each link per each time step
    energy_flow = {
        t: {out: RealVariable() for out in range(len(connections))}
        for t in hubs[0].time
    }

    # Constructing list of energy flows for the links
    for i in range(len(connections)):
        hub = hubs[0]
        flow = []
        for t in hub.time:
            flow.append(energy_flow[t][i])

        links.insert(i, flow)

    # Creating a constraint that sets the bounds of the flow in the links:
    # link[i] >= 0. To ensure one directional connection
    # link[i] <= link_capacity.
    for hub in hubs:
        for i, link in enumerate(links):
            for c in link_capacity_constraint(link, hub, i):
                constraints.append(c)

    LINK_THERMAL_LOSS = _net_data.link_thermal_loss

    # Create a network connection
    for k, hub in enumerate(hubs):
        # New list for all the links that the hub starts at
        link_start = []
        # New list for all the links that the hub ends at
        link_end = []

        for i in range(len(connections)):
            # Searching for the links that the hub starts at
            if connections[i][0] == k:
                link_start.append(links[i])
            # Searching for the links that the hubs ends at
            if connections[i][1] == k:
                # Multiplying by thermal loss to account for the thermal loss in the link
                links[i] = np.array(links[i])
                links[i] = links[i] * LINK_THERMAL_LOSS[i]
                link_end.append(links[i])

        # Creating the network constraints for the hubs
        for c in network_constraint(hub, link_end, link_start):
            constraints.append(c)

    # The objective function of the model is the summation of all the hub's
    # objective function
    objective = hubs[0].objective

    for hub in hubs[1:]:
        objective += hub.objective

    # Now solve this model.
    status = pylp.solve(objective=objective, constraints=constraints, minimize=True)

    #: Define the different sheets that we output
    sheets = [
        "Other",
   	    "LOADS",
   	    "SOLAR_EM",
   	    "energy_exported",
   	    "energy_from_storage",
   	    "energy_imported",
   	    "energy_to_storage",
   	    "is_on",
   	    "storage_level",
        "capacity_storage",
        "capacity_tech",
        ]

    #: Used to output files with names according to their job.
    if minimize_carbon:
        ext = "_minimized_carbon"
    else:
        ext = "_minimized_cost"

    #: A list used to return all dictionaries gathered from this function.
    hub_dict_list = []

    # Print out the hub's name and all stuff related to that model.
    tcar = 0
    tcost = 0
    tnetwork = 0
    for i, hub in enumerate(hubs):
        hub_dict_list.append(hub.solution_dict)
        if output_filename is None:
            output_excel(hub.solution_dict, f'{hub.name}{ext}.xlsx', sheets=sheets)
            print_section(f'{hub.name}{ext}', hub.solution_dict)
        else:
            output_excel(hub.solution_dict, f'{output_filename}_{i}{ext}.xlsx', sheets=sheets)
            print_section(f'{hub.name}_{i}{ext}', hub.solution_dict)
        tcar += hub.solution_dict["total_carbon"]
        tcost += hub.solution_dict["total_cost"]
        tnetwork += hub.solution_dict["network_cost"]

    for i, hub in enumerate(hubs):
        half_heading = '=' * 4
        for t in hub.time:
            net_imp = hub.solution_dict['energy_imported'][t]['Net_import']
            net_exp = hub.solution_dict['energy_exported'][t]['Net_export']

            if (net_imp != 0) and (net_exp != 0):
                print(f"{half_heading} {'Warning:'} {half_heading} \n{'Export & import at the same time for hub: '} {i}"
                      f"{'; For time step: '} {t} ")

    print("\nAbsolute total carbon:\t" + str(tcar))
    print("Absolute total cost:\t" + str(tcost), end='\n\n')
    for hub in hubs:
        print(hub.name + " network cost: " + str(hub.solution_dict['network_cost']))
    print("\nAbsolute total network cost:\t" + str(tnetwork))
    return hub_dict_list


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="Command line utility to run tests.")
    parser.add_argument('-c', '--carbon', action='store_true', help='Optimize based on the total carbon level.')
    parser.add_argument('-o', '--output_filename', type=str, help='The name that the output should be stored under.')
    parser.add_argument('-v', '--carbon_value', type=float, help='The value that the carbon should try to be.')
    parser.add_argument('-n', '--number_of_hubs', type=int, help='The number of hubs to be used')
    parser.add_argument('-i', '--input_files', type=str, help='The name of the directory and the beginning of the'
                                                              'the input files ')
    parser.add_argument('--network', type=str, help='The name that the network excel file')

    args = parser.parse_args()

    multiple_hubs(args.carbon, args.output_filename, args.input_files, args.network,
                  args.carbon_value, args.number_of_hubs)
